# Auditory stimulus:
#   - continuous 40 Hz AM tone with f0 = 500 Hz
# Created using 2017 version of mn.py library by Mats Nilsson, 
# see the newest version on https://github.com/stamnosslin/mn

# Visual stimuli: 
#   - upright and inverted crosses in five colors

# Two visual task loads: 
#   - low load: detect red colored crosses
#   - high load: detect upright yellow cross and green inverted cross

# each condition is repeated 4 times (together: 8 blocks)

# Randomizations: 
#   - LL and HL alternate, the initial load determined randomy for each subject
#   - color-orientation combinations are presented equally often, 
#       in pseudo random order
#   - there can be between 2 and 6 consecutive non-targets
#   - 20% of trials = targets

from __future__ import division
from psychopy import visual
from psychopy import core, gui, data, event, parallel
from psychopy.tools.filetools import fromFile, toFile
import sounddevice as sd
import time, numpy, random, os.path, math, sys
import numpy as np
import mn_mmn4 as mn
import matplotlib.pyplot as plt
import operator

# ==============================================================================
# BASIC SETTINGS ---------------------------------------------------------------

p_port = parallel.ParallelPort(address = "0x4020")
#os.chdir('C:\Users\MASZ6827\Documents\PhD_studies\_Attention_ASSRs\experiment\data')
#os.chdir('E:\Malinowe\ASSR_study1\experiment\data')
#os.chdir('C:\Users\SWIENSLAB03\Box Sync\ASSR_study1\experiment\data') #box in the lab
os.chdir('C:\Users\SWIENSLAB03\Desktop\ASSR1\experiment\data') #desktop in the lab

# Set the experimental parameters ----------------------------------------------
random.seed()
nTrialsPerBlock = 360
nCrosses = 10
nTargets = 2
nNontargets = 8
nBlocks = 8
nTrials = 2936
TrialPerCrossPerBlock = int(nTrialsPerBlock/nCrosses)

# Present a dialogue to give subject nr and gain:
expInfo = gui.Dlg(title="ASSRs_1")
expInfo.addField('Subject', '999')
expInfo.addField('Threshold', choices = [15, 10, 5, 0, -5, -10, -15, -20])
ok_data = expInfo.show()  # show dialog and wait for OK or Cancel

if expInfo.OK:  # or if ok_data is not None
    print('Subject: %i\n' %(int(ok_data[0])))
else:
    print('user cancelled')

# Quit if sound level too high:
#if float(ok_data[1]) > -10:
#    print('Error: gain > -10') 
#    core.quit()

# Screen:
win = visual.Window([1920, 1080], allowGUI = False, # [1920, 1080]
    monitor = 'testMonitor', units = 'height', color = 'gray')
    
    
# Make a text file to save data ------------------------------------------------
fileName = 'fp_' + ok_data[0]
dataFile = open(fileName+'.txt', 'w')
dataFile.write('fp\tblock_nr\ttrial_nr\tcondition\tsound\ttrigger\ttarget\tcross_cond\ttrial_start\tresponse_time\n')


# ==============================================================================
# FUNCTIONS --------------------------------------------------------------------

def check4esc():
    ''' Press ESC to force stop the experiment '''
    keys = event.getKeys(keyList = ['escape'])
    if len(keys)>0 and 'escape' in keys:
        core.quit()

def trigger(trig,dur):
    p_port.setData(trig)
    core.wait(dur) # duration of trigger
    p_port.setData(255) # turn sound trigger off

def paus(t,pauses,block,trig = 223):
    ''' 
    Keyword arguments:
    pauses -- list of trialnumbers where there should be a pauses (list)
    t -- the current trialnumber (int)
    block -- list of blocknumbers, ex (0,0,0,1,1,1,2,2,2) (list)
    trig -- what trigger to send for pause (int)
    '''
    if t in pauses:
        if t != pauses[len(pauses)-1]:
            paus_info.setText('Short break\n%i/%i blocks done.\nHit space when ready.'% (block[t+1], len(pauses)))
            paus_info.draw()
            win.flip()
            event.waitKeys(keyList=['space'])
            trigger(trig,0.01)
            win.flip()
            core.wait(2)

def response_check(key):
    """
    Checks if a key was pressed.
    Keyword arguments:
    key -- containing either a keypress and a time or nothing (list)

    return:
    time -- nan if not pressed or time of press if pressed
    """
    if len(key) == 0:
        pressed = 0
    elif 'space' in key[0]:
        pressed = 1
    if pressed == 1:
        time = key[0][1]
    elif pressed == 0:
        time = 'nan'

    return str(time), pressed

# Assign crosses to response and nonresponse trials
def cross(responses,targets,nontargets):
    '''
    Assigns cross conditions (colors * 5 and orientations * 2) to response and 
    nonresponse trials for high and low load
    
    Keyword arguments:
    reponses -- list of 1 and 0 for target and nontarger respectively
    targets -- list target condtions (see above)
    nontargets -- list of nontarget conditions
    
    Return:
    cross_cond -- a list of conditions (0 to 9)
    '''
    T_Pool = []
    NT_Pool  = []
    for b in range(nBlocks):
        target_pool = []
        nontarget_pool = []
        crossesAdd7 = list(np.random.choice(nontargets, 7, replace = False))
        for t in range(TrialPerCrossPerBlock ): # creates target codition pools per block
            target_pool += targets # target pool
        for t in range(TrialPerCrossPerBlock ): # creates nontarget codition pools per block
            nontarget_pool += nontargets # nontarget pool
        T_PoolPerBlock = list(np.random.choice(target_pool, (TrialPerCrossPerBlock )*nTargets, replace = False)) # randomizes the pools
        NT_PoolPerBlock = list(np.random.choice(nontarget_pool, (TrialPerCrossPerBlock )*nNontargets, replace = False))
        NT_PoolPerBlock = crossesAdd7 + NT_PoolPerBlock
        
        T_Pool.append(T_PoolPerBlock)
        NT_Pool.append(NT_PoolPerBlock)
    #T_Pool = reduce(operator.concat, T_Pool)
    #NT_Pool = reduce(operator.concat, NT_Pool)
    #print(T_Pool)
    
    cross_cond = []
    for b in range (nBlocks):
        cross_cond += NT_Pool[b][0:7] #append first 7 non target trials
        # Go into "responses" but omit first 7 trial per each block 
        # and assign crosses
        for i in responses[b][7:len(responses[b])]: 
            if i == 1: # randomly assigns a target cross to a response trial
                index = np.random.choice(range(len(T_Pool[b])), 1) # randomly select a position from the pool
                cross_cond += [T_Pool[b][index[0]]] # assign randomized index
                del T_Pool[b][index[0]] # delete the selected one from the pool
            elif i == 0: # randomly assigns a target cross to a nonresponse trial
                index = np.random.choice(range(7, len(NT_Pool[b])), 1) # note: we want to omit the first 7 trials in each block
                cross_cond += [NT_Pool[b][index[0]]]
                del NT_Pool[b][index[0]]
    return cross_cond


def show_target_crosses(pauses, b, block_paradigm_current, targethi, targetlo):
    '''
    Shows what crosses are targets for this block.
    
    Keyword arguments:
    pauses -- list of trial numbers where there should be a pauses (list)
    t -- the current trial number (int) 
    '''
    # start_block = [i - 366 for i in pauses]
    # if t in start_block:
    lohi = targetlo,targethi
    #set stimuli according to condition
    t1_standing = visual.Rect(win=win, name='up_cross_hor', width=(dimentions[1]),
        height=(dimentions[0]), ori=0, 
        pos=(-0.2, 0-0.2), lineWidth=1, 
        lineColor=colors[lohi[block_paradigm[b]-1][0]], lineColorSpace='rgb', 
        fillColor=colors[lohi[block_paradigm[b]-1][0]], fillColorSpace='rgb', opacity=1, depth=0.0, interpolate=True)
    
    t1_laying = visual.Rect(win=win, name='up_cross_hor', width=(dimentions[0]), 
        height=(dimentions[1]), ori=0, 
        pos=(-0.2, position[lohi[block_paradigm[b]-1][0]]-0.2), lineWidth=1, 
        lineColor=colors[lohi[block_paradigm[b]-1][0]], lineColorSpace='rgb', 
        fillColor=colors[lohi[block_paradigm[b]-1][0]], fillColorSpace='rgb', opacity=1, depth=0.0, interpolate=True)
    
    t2_standing = visual.Rect(win=win, name='up_cross_hor', width=(dimentions[1]), 
        height=(dimentions[0]), ori=0, 
        pos=(0.2, 0-0.2), lineWidth=1, 
        lineColor=colors[lohi[block_paradigm[b]-1][1]], lineColorSpace='rgb', 
        fillColor=colors[lohi[block_paradigm[b]-1][1]], fillColorSpace='rgb', opacity=1, depth=0.0, interpolate=True)
    
    t2_laying = visual.Rect(win=win, name='up_cross_hor', width=(dimentions[0]), 
        height=(dimentions[1]), ori=0, 
        pos=(0.2, position[lohi[block_paradigm[b]-1][1]]-0.2), lineWidth=1, 
        lineColor=colors[lohi[block_paradigm[b]-1][1]], lineColorSpace='rgb', 
        fillColor=colors[lohi[block_paradigm[b]-1][1]], fillColorSpace='rgb', opacity=1, depth=0.0, interpolate=True)
            
    t1_standing.draw() # show vertical bar
    t1_laying.draw() # show horizontal bar
    t2_standing.draw() # show vertical bar
    t2_laying.draw() # show horizontal bar
    target_cross_info.draw()
    win.flip()
    event.waitKeys(keyList=['space'])
    win.flip()
    core.wait(2)
    

# ==============================================================================
# CREATE VISUALS ---------------------------------------------------------------

intro_text = visual.TextStim(
    win, color = 'white', height = 0.03, pos = (0, 0), text = 'Press space to start the experiment!')
end_text = visual.TextStim(
    win, color = 'white', height = 0.03, pos = (0, 0), text = 'End of experiment!\n\nThank you for participating!')
paus_info = visual.TextStim(
    win, color = 'white', height = 0.03, pos = (0, 0), text = 'paus')
processing = visual.TextStim(
    win, color = 'white', height = 0.03, pos = (0, 0), text = 'Setting up experiment.')
target_cross_info = visual.TextStim(
    win, color = 'white', height = 0.03, pos = (0, 0), text = 'Your task is to press space when you see these crosses. \nTo continue to the task, press the spacebar.')

#photo diode stimulator
whiteOn = visual.Rect(
    win=win, name='WhiteOn',
    width=(0.025, 0.025)[0], height=(0.025, 0.025)[1],
    ori=0, pos=(0.88, 0.49),
    lineWidth=1, lineColor=[1.000,1.000,1.000], lineColorSpace='rgb',
    fillColor=[1.000,1.000,1.000], fillColorSpace='rgb',
    opacity=1, depth=0.0, interpolate=True)	

processing.draw()
win.flip()


# ==============================================================================
# CREATE SOUNDS ----------------------------------------------------------------

sd.default.device = "ASIO Fireface USB"
print 'Sound device ----------------------------------------------------------'
print sd.default.device #device = "ASIO Fireface USB")
print '-----------------------------------------------------------------------'

# Gain    -->  SPL  SL
# -15.7   -->  75   +15
# -20.5   -->  70   +10
# -25.3   -->  65   +5 
# -30.1   -->  60    0 
# -34.8   -->  55   -5
# -39.6   -->  50   -10
# -44.4   -->  45   -15
# -49.1   -->  40   -20

# Set the gain and sampling frequency (fs)
thresholds = [15, 10, 5, 0, -5, -10, -15, -20]
level_pool = [-15.7, -20.5, -25.3, -30.1, -34.8, -39.6, -44.4, -49.1]
# equivalent [ 75,    70,    65,    60,    55,    50,    45,    40] dB SPL

for i in range(8):
    if float(ok_data[1]) == thresholds[i]:
        gain	= level_pool[i] # -30.1, To set up with each participant
print(gain)

fs 		= 44100		# Sampling frequency
f0 		= 500		# Fundamental frequency
fAM 	= 40.96		# Modulation frequency

# Create fundamental frequency sine
sine_f0 	= mn.create_sinusoid(freq = f0, phase = 0, fs = fs, dur = 200) #dur in sec 
sine_f0_inv = mn.create_sinusoid(freq = f0, phase = math.pi, fs = fs, dur = 200) #dur in sec 	
# Create modulation frequency sine
sine_fAM 	= mn.create_sinusoid(freq = fAM, phase = -math.pi/2, fs = fs, dur = 200) # shift in phase to cosine, to start the wave from 0
# Multiply the two frequencies to create amplitude modulated signal
sine_fAM    = sine_fAM + 1
AMtone 		= sine_f0 * sine_fAM 
AMtone_inv  = sine_f0_inv * sine_fAM 
# Apply fade (10 ms) and gain
AMtone 		= mn.set_gain(mn.fade(AMtone,int(fs/100)),gain) # fs/100 gives 10 ms
AMtone_inv	= mn.set_gain(mn.fade(AMtone_inv,int(fs/100)),gain) # fs/100 gives 10 ms
# Transpose and combine the signal to create two channels audio array
AMtone 		= np.transpose(np.array([AMtone,AMtone]))
AMtone_inv 	= np.transpose(np.array([AMtone_inv,AMtone_inv]))

tones       = [AMtone, AMtone_inv]
tones_names  = ['40Hz_AM', '40Hz_AM_inv']
## Quick graphical check:
# s = AMtone 
# plt.plot(s)
# plt.xlabel('samples')
# plt.ylabel('voltage (mV)')
# plt.grid(True)
# plt.show()


# ==============================================================================
# MAKE BLOCKS, PAUSES AND CONDITIONS

blocks = []
pauses = []
for i in range(nBlocks):
    pauses.append([])
    pauses[i] = int(nTrials/nBlocks * i + (nTrials/nBlocks - 1))
    blocks += [i] # * int(nTrials/nBlocks)

print(blocks)
# ==============================================================================
# CREATE TRIAL ORDER
# Randomized in a way that there are small blocks of 5 items. 
# The target is always somewhere among the first 3 items. 
# There is always at least 2 non-targets between the consecutive targets
# It refers to previously defined function: cross(responses, targets, nontargets)
# Note: This creates responses in low load and high load condition throught 
# the whole experiment. Later in the procedure it selects from which list 
# it should take the trial
 
'''
nTrialsPerBlock = 360
nCrosses = 10
nTargets = 2
nNontargets = 8
nBlocks = 8
'''

# Cross conditions:
#                                low load                       high load
# up    red     0   low load    o x o o o o --> 253     o o x o o o --> 251
# down  red     1   low load    o x o o o o --> 253     o o x o o o --> 251
# up    blue    2               x o o o o o --> 254     o o x o o o --> 251
# down  blue    3               x o o o o o --> 254     o o x o o o --> 251
# up    yellow  4   hi load     x o o o o o --> 254     o o o x o o --> 247
# down  yellow  5               x o o o o o --> 254     o o x o o o --> 251
# up    purple  6               x o o o o o --> 254     o o x o o o --> 251
# down  purple  7               x o o o o o --> 254     o o x o o o --> 251
# up    green   8               x o o o o o --> 254     o o x o o o --> 251
# down  green   9   hi load     x o o o o o --> 254     o o o x o o --> 247
# space press                   o o o o o x --> 223
# beginning of the block        x x x x x x --> 192

crossesLT   = [0, 1] # low load targets crosses
crossesHT   = [4, 9] # high load targets crosses
crossesLNT  = [2, 3, 4, 5, 6, 7, 8, 9] # low load no targets crosses
crossesHNT  = [0, 1, 2, 3, 5, 6, 7, 8] # high load no targets crosses

targets_all = []

for b in range(nBlocks): 
    targets_block = [[0]] * 7
    for t in range(72): # because we have 360 trials / 5 items (4 non targets and 1 target)
        possible_targets    = list(np.random.choice([1,0,0], 3, replace = False))
        two_zeros           = [0, 0]
        targets_block.append(possible_targets)
        targets_block.append(two_zeros)
    targets_block = reduce(operator.concat, targets_block) # this reduces list of list to a flat list
    targets_all.append(targets_block)

targets_all_onelist = reduce(operator.concat, targets_all)

# Assign crosses to response and nonresponse trials ----------------------------
LL_trialOrder = cross(responses = targets_all, targets = crossesLT, nontargets = crossesLNT)
HL_trialOrder = cross(responses = targets_all, targets = crossesHT, nontargets = crossesHNT)

all_paradigms       = [LL_trialOrder, HL_trialOrder]
#                       low load                                  high load                              
#                       T,  T,  NT, NT, NT, NT, NT, NT, NT, NT,   NT, NT, NT, NT, T,  NT, NT, NT, NT, T  
trig_list           = [[253,253,254,254,254,254,254,254,254,254],[251,251,251,251,247,251,251,251,251,247]] 
# [[248,216,254,222,253,221,251,219,247,215],[240,208,252,220,249,217,243,211,231,199]]
all_responses       = [targets_all_onelist, targets_all_onelist]
all_crosses         = [LL_trialOrder, HL_trialOrder]

# block_paradigm:
# 1 - low load
# 2 - high load 

np.random.seed(int(ok_data[0]))
block_paradigm = [[1,2,1,2,1,2,1,2],[2,1,2,1,2,1,2,1]]
block_paradigm = block_paradigm[int(np.random.choice([0,1], 1))] # maybe should be counterbalanced??
tone_paradigm  = [[0,0,1,1,0,0,1,1],[1,1,0,0,1,1,0,0]]
tone_paradigm  = tone_paradigm[int(np.random.choice([0,1], 1))]
np.random.seed()

print(block_paradigm)
print(tone_paradigm) 
    
trial_paradigm = [] # Contains the paradigm on every trial (1 to 4)
for i in range(nBlocks):
    trial_paradigm += [block_paradigm[i]]*367 # 360 trials per condition(block) plus 7 standards in beginning of each one

#           red                 blue              yellow            purple             green
colors = [1,-1,-1],[1,-1,-1],[-1,-1,1],[-1,-1,1],[1,1,-1],[1,1,-1],[1,-1,1],[1,-1,1],[-1,1,-1],[-1,1,-1]
# ver, hor
dimentions = [0.100,0.010]
position = [0.0255,-0.0255]*5

win.flip()
globalClock = core.Clock()  # to keep track of time
respClock = core.Clock()

# ==============================================================================
# Experimental procedure -------------------------------------------------------
trigger(255,0.01) # Turn off all triggers
intro_text.draw()
win.flip()
event.waitKeys(keyList=['space'])
win.flip()

for b in range(nBlocks):
    print block_paradigm[b]
    # show target for next block
    show_target_crosses(pauses, b, block_paradigm[b], crossesHT, crossesLT)
    trigger(192,0.01) # Signalize beginning of the block
    sd.play(tones[tone_paradigm[b]], fs) # Play sound
    core.wait(0.2)
    for t in range(367): # this is the total amount of trials
        l_trial_start = globalClock.getTime()
        check4esc() # check for esc

        #set stimuli according to condition
        standing = visual.Rect(win=win, name='up_cross_hor', width=(dimentions[1]), 
            height=(dimentions[0]), ori=0, pos=(0, 0), lineWidth=1, 
            lineColor=colors[all_crosses[block_paradigm[b]-1][t]], 
            lineColorSpace='rgb', fillColor=colors[all_crosses[block_paradigm[b]-1][t]], 
            fillColorSpace='rgb', opacity=1, depth=0.0, interpolate=True)
        laying = visual.Rect(win=win, name='up_cross_hor', width=(dimentions[0]), 
            height=(dimentions[1]), ori=0, pos=(0, position[all_crosses[block_paradigm[b]-1][t]]), lineWidth=1, 
            lineColor=colors[all_crosses[block_paradigm[b]-1][t]], 
            lineColorSpace='rgb', fillColor=colors[all_crosses[block_paradigm[b]-1][t]], 
            fillColorSpace='rgb', opacity=1, depth=0.0, interpolate=True)

        trigger(trig_list[block_paradigm[b]-1][all_paradigms[block_paradigm[b]-1][t]],0.01) # send condition trigger

        standing.draw() # vertical bar
        laying.draw() # horizontal bar
        whiteOn.draw() # square

        win.flip() # show cross and white
        #key = [] #not necessary
        #key = event.getKeys(keyList = ['space'], timeStamped = globalClock) # not necessary
        core.wait(0.089, hogCPUperiod = 0.089)
        win.flip() # turn visual stuff off
        #if response_check(key)[1] == 1:    # send the trigger when button is pressed
        #    trigger(223,0.01)              # doesnt work as it should
        core.wait(0.381, hogCPUperiod = 0.381) # ITI

        l_fp = int(ok_data[0])
        l_block_nr = blocks[b]+1
        l_trial_nr = (range(367)*nBlocks)[t]+1
        l_condition = block_paradigm[b]
        l_sound = tones_names[tone_paradigm[b]]
        l_trigger = trig_list[block_paradigm[b]-1][all_paradigms[block_paradigm[b]-1][t]]
        l_target_cross = all_responses[block_paradigm[b]-1][t]
        l_cross_condition = all_crosses[block_paradigm[b]-1][t]
        #l_trial_start = trial_start
        key = event.getKeys(keyList = ['space'], timeStamped = globalClock)
        #if response_check(key)[1] == 1:    # send the trigger when button is pressed
        #    trigger(223,0.01)              # doesnt work as it should
        l_response_time = response_check(key)[0]
        # Save data to file
        #'fp\tblock_nr\ttrial_nr\tcondition\tsound\ttrigger\ttarget_cross\tcross_cond\ttrial_start\tresponse_time\n'
        dataFile.write('%i\t%i\t%i\t%i\t%s\t%i\t%i\t%i\t%f\t%s\n' %(
            l_fp, l_block_nr, l_trial_nr, l_condition, l_sound, l_trigger, 
            l_target_cross, l_cross_condition, l_trial_start, l_response_time))

    sd.stop()
    paus_info.setText('Short break\n%i/%i blocks done.\nHit space when ready.'% (b+1, len(pauses)))
    paus_info.draw()
    win.flip()
    event.waitKeys(keyList=['space'])
    win.flip()
    core.wait(2)
    
dataFile.close()
end_text.draw()
win.flip()
event.waitKeys(keyList=['space'])

